import numpy as np
import pandas as pd
import time
import argparse
from .ordinal_patterns_calculation import calculate_ordinal_patterns
from .entropy_calculation import sliding_window_entropy, linear_normalization, range_normalization, default_normalization


def create_filenames(input_filename, timelag, gap, window_size):
    k = input_filename.rfind('.')
    entropy_filename = input_filename[:k] + '_entropy_' + str(window_size) + '_' + str(timelag) + '_' + str(gap) + input_filename[k:]
    ordinal_filename = input_filename[:k] + '_ordinal_' + str(timelag) + '_' + str(gap) + input_filename[k:]
    shortened_filename = input_filename[:k] + '_shortened_' + str(timelag) + '_' + str(gap) + input_filename[k:]
    return shortened_filename, ordinal_filename, entropy_filename


def process_data(input_filename, timelag=5, gap=0, window_size=512, normalization=range_normalization, verbose=False):
    shortened_filename, ordinal_filename, entropy_filename = create_filenames(input_filename, timelag, gap, window_size)
    data_df = pd.read_csv(input_filename)

    start_time = time.clock()
    shortened_df, ordinal_df = calculate_ordinal_patterns(data_df, timelag, gap)
    end_time = time.clock()

    if verbose:
        print('bandt_pompe took ', str(end_time - start_time), 's')

    shortened_df.to_csv(shortened_filename, index=False)
    if verbose:
        print('shortened data saved to ', shortened_filename)

    ordinal_df.to_csv(ordinal_filename, index=False)
    if verbose:
        print('ordinal pattern representation saved to ', ordinal_filename)

    start_time = time.clock()
    entropy_df = sliding_window_entropy(ordinal_df, window_size, pattern_length=timelag, normalization=normalization)
    end_time = time.clock()

    if verbose:
        print('entropy took ', str(end_time - start_time), 's')

    entropy_df.to_csv(entropy_filename, index=False)
    if verbose:
        print('entropy saved to ', entropy_filename)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument('file',  type=str, help="source file that contains the raw data (should be csv)")
    parser.add_argument('timelag', type=int, help="timelag, which determines the length of the ordinal patterns")
    parser.add_argument('gap', type=int, help="gap between subsequent values that are used for creating ordinal patterns")
    parser.add_argument('windowsize', type=int, help="windowsize for calculation entropy based on sliding window")

    norm_group = parser.add_mutually_exclusive_group()
    norm_group.add_argument('-linear', '-l', action='store_true', help='use linear normalization for entropy')
    norm_group.add_argument('-range', '-r', action='store_true', help='use range normailzation to interval [0, 1] for entropy (default)')

    parser.add_argument('--verbosity', '-v', action='store_true', help="display additional details like runtime and savend output files")

    args = parser.parse_args()

    filename = args.file
    gap = args.gap
    timelag = args.timelag
    window_size = args.windowsize
    verbosity = args.verbosity

    if args.linear:
        normalization = linear_normalization
    else:
        normalization = range_normalization

    process_data(filename, timelag, gap, window_size, normalization, verbosity)
