import numpy as np
import pandas as pd
from bandt_pompe.ordinal_patterns_calculation import calculate_ordinal_patterns
from os.path import exists
from os import makedirs
from json import dumps
from .feature_extraction import AbstractFeatureExtraction
from scipy.stats import norm


class DataProcessing:

    def __init__(self, filename, degree, gap, experiment):
        self.degree = degree
        self.gap = gap

        pos_slash = filename.rfind('/')
        pos_dot = filename.rfind('.')
        self.dirname = filename[0:pos_slash+1]+filename[pos_slash+1:pos_dot]+'_'+str(self.degree)+'_'+str(self.gap)
        self.filename = filename
        self.experiment = experiment

    def _calculate_ordinal_patterns(self):
        min_size = (self.degree - 1) * self.gap + self.degree
        df = pd.read_csv(self.filename)

        to_concat_short = []
        to_concat_ordinal = []
        for current_experiment in df[self.experiment].unique():
            filter = (df[self.experiment]==current_experiment)
            current_df = df.loc[filter, :].reset_index(drop=True)
            if len(current_df) >= min_size:
                short_temp, ordinal_temp = calculate_ordinal_patterns(current_df, self.degree, self.gap)
                to_concat_short.append(short_temp)
                to_concat_ordinal.append(ordinal_temp)

        ordinal_df = pd.concat(to_concat_ordinal, ignore_index=True)
        short_df = pd.concat(to_concat_short, ignore_index=True)

        ordinal_df[self.experiment] = short_df[self.experiment]
        return short_df, ordinal_df

    def get_ordinal_data(self, target=None, recalc=False, dump=True):
        short_filename = self.dirname + '/' + 'short.csv'
        ordinal_filename = self.dirname + '/' + 'ordinal.csv'

        if not exists(short_filename) or not exists(ordinal_filename) or recalc:
            short_df, ordinal_df = self._calculate_ordinal_patterns()

            if dump:
                if not exists(self.dirname):
                    makedirs(self.dirname)
                short_df.to_csv(short_filename, index=False)
                ordinal_df.to_csv(ordinal_filename, index=False)

        else:
            short_df = pd.read_csv(short_filename)
            ordinal_df = pd.read_csv(ordinal_filename)

        if target is not None:
            ordinal_df[target] = short_df[target]

        return short_df, ordinal_df

    def get_extracted_data(self, feature_extraction: AbstractFeatureExtraction,
                           target=None, fit=True, recalc=False, dump=True):
        extracted_filename = self.dirname + '/' + 'extracted' + '_' + feature_extraction.get_name() + '.csv'

        if not exists(extracted_filename) or recalc:
            short, ordinal = self.get_ordinal_data(target, dump=dump)

            if fit:
                feature_extraction.fit(ordinal, target, self.experiment)
            extracted = feature_extraction.transform(ordinal, self.experiment)

            if target is not None:
                extracted[target] = feature_extraction.transform_target(short, target, self.experiment)

            if dump:
                if not exists(self.dirname):
                    makedirs(dirname)
                extracted.to_csv(extracted_filename, index=False)

        else:
            extracted = pd.read_csv(extracted_filename)

        return extracted


class SAX:

    def __init__(self, number, mode='uniform'):
        self.number = number
        self.mode = mode
        self.bins = {}

    def fit(self, data):
        self.bins = {}
        for col, values in data.iteritems():
            if self.mode == 'normal':
                calc_bins = np.vectorize(lambda x: norm.ppf(x, values.mean(), values.std()))
                self.bins[col] = calc_bins(np.arange(0, 1, 1/self.number))
            else:
                self.bins[col] = np.arange(values.min(), values.max(), (values.max() - values.min())/self.number)

    def transform(self, data):
        result = {}
        for col, bins in self.bins.items():
            result[col] = np.digitize(data[col], bins).astype(str)

        return pd.DataFrame(result)
