import numpy as np
import pandas as pd
import json
from random import randint, shuffle, uniform
from math import ceil, pow, exp
from .scoring import experimentwise_probabilities, classwise_mean, \
                     classwise_variance, build_chebyshev_matrix, \
                     get_slice_filter, JSD, KLD, MI
from scipy.stats import spearmanr, pearsonr
from sklearn.metrics import mutual_info_score


class AbstractFeatureExtraction():

    def transform_target(self, data, target, experiment):
        return np.array(data[target].groupby(data[experiment]).first())

    def get_name(self):
        raise NotImplementedError()

    def transform(self, data, target, experiment):
        raise NotImplementedError()

    def fit(self, data, experiment):
        raise NotImplementedError()


class IncrementalFeatureExtraction(AbstractFeatureExtraction):

    def __init__(self, feature_number, alpha, max_subspace_size):
        self.alpha = alpha
        self.feature_number = feature_number
        self.max_subspace_size = max_subspace_size
        self.slices = []
        self.cached_values = {}
        self.score = 0
        self.redundancies = np.ma.array(np.zeros((self.feature_number, self.feature_number)),
                                                  mask = np.identity(self.feature_number))
        self.relevance = 0

    def _get_values(self, data, feature):
        if feature not in self.cached_values:
            values, counts = np.unique(data[feature], return_counts=True)
            self.cached_values[feature] = pd.Series(index=values, data=counts)
        return self.cached_values[feature]

    def _generate_slice(self, data, subspace):
        subspace.sort()
        new_slice = {'score': 0, 'feature_set': subspace, 'conditions': {}}
        samples_per_dimension = ceil(pow(self.alpha, 1/len(subspace)) * len(data))

        for ft in subspace:
            ft_values = self._get_values(data, ft)
            shuffled_values = np.random.permutation(ft_values.index.values)
            sample_counter = 0
            condition = []

            for value in shuffled_values:
                if sample_counter >= samples_per_dimension:
                    break
                condition.append(value)
                sample_counter += ft_values[value]
            new_slice['conditions'][ft] = condition

        return new_slice

    def get_name(self):
        name = 'none_redundant_' + str(self.feature_number) + '_' + str(self.alpha) + '_' + str(self.max_subspace_size)
        return name

    def _maintain_slices(self, generated_slices, data, target, experiment):
        experiment_group = data.groupby(target)[experiment]
        for s1 in generated_slices:

            slice_filter = get_slice_filter(data, s1)

            probabilities = experimentwise_probabilities(slice_filter, data, experiment)
            means = classwise_mean(experiment_group, probabilities)
            variances = classwise_variance(experiment_group, probabilities)
            relevance = build_chebyshev_matrix(means, variances).min()

            s1['probabilities'] = probabilities
            s1['relevance'] = relevance
            s1_redundancies = np.zeros(self.feature_number)

            if np.isinf(s1['relevance']):
                break

            for i in range(len(self.slices)):
                s2 = self.slices[i]
                s1_redundancies[i] = abs(spearmanr(s1['probabilities'], s2['probabilities'])[0])

            # simply add, with feature_number isn't reached yet
            if len(self.slices) < self.feature_number:
                self.redundancies[len(self.slices), :] = s1_redundancies
                self.redundancies[:, len(self.slices)] = s1_redundancies
                self.redundancies.mask = np.identity(self.feature_number)
                self.relevance += s1['relevance']
                self.slices.append(s1)

            # test if slice is beneficial
            else:
                max_score = self.score
                changed = -1
                for i in range(len(self.slices)):
                    red_copy = self.redundancies.copy()
                    red_copy[i, :] = s1_redundancies
                    red_copy[:, i] = s1_redundancies
                    red_copy.mask = np.identity(self.feature_number)
                    red = red_copy.max(axis=1).sum()

                    rel = self.relevance - self.slices[i]['relevance'] + s1['relevance']
                    if max_score < rel-red:
                        max_score = rel-red
                        changed = i

                if changed >= 0:
                    self.redundancies[changed, :] = s1_redundancies
                    self.redundancies[:, changed] = s1_redundancies
                    self.redundancies.mask = np.identity(self.feature_number)
                    self.relevance -= self.slices[changed]['relevance']
                    self.relevance += s1['relevance']
                    self.slices[changed] = s1

            # update score
            self.score = self.relevance-self.redundancies.max(axis=1).sum()

    def fit(self, data, target, experiment, iterations=2000):
        input_fts = [ft for ft in data.columns.values if ft != target and ft != experiment]
        generated_slices = []
        for i in range(iterations):
            subspace_size = randint(1, self.max_subspace_size)
            shuffle(input_fts)
            subspace = input_fts[:subspace_size]

            s = self._generate_slice(data, subspace)
            generated_slices.append(s)

        self._maintain_slices(generated_slices, data, target, experiment)

    def transform(self, data, experiment):
        experiments = data[experiment].unique
        new_features = {}

        for i in range(len(self.slices)):
            s = self.slices[i]
            slice_filter = get_slice_filter(data, s)
            name = json.dumps(s['feature_set']) + str(i)
            new_feature = np.array(experimentwise_probabilities(slice_filter, data, experiment))
            new_features[name] = new_feature

        return pd.DataFrame(data=new_features)


class IncrementalFeatureExtractionPearsonRedundancy(AbstractFeatureExtraction):

    def __init__(self, feature_number, alpha, max_subspace_size):
        self.alpha = alpha
        self.feature_number = feature_number
        self.max_subspace_size = max_subspace_size
        self.slices = []
        self.cached_values = {}
        self.score = 0
        self.redundancies = np.ma.array(np.zeros((self.feature_number, self.feature_number)),
                                                  mask = np.identity(self.feature_number))
        self.relevance = 0

    def _get_values(self, data, feature):
        if feature not in self.cached_values:
            values, counts = np.unique(data[feature], return_counts=True)
            self.cached_values[feature] = pd.Series(index=values, data=counts)
        return self.cached_values[feature]

    def _generate_slice(self, data, subspace):
        subspace.sort()
        new_slice = {'score': 0, 'feature_set': subspace, 'conditions': {}}
        samples_per_dimension = ceil(pow(self.alpha, 1/len(subspace)) * len(data))

        for ft in subspace:
            ft_values = self._get_values(data, ft)
            shuffled_values = np.random.permutation(ft_values.index.values)
            sample_counter = 0
            condition = []

            for value in shuffled_values:
                if sample_counter >= samples_per_dimension:
                    break
                condition.append(value)
                sample_counter += ft_values[value]
            new_slice['conditions'][ft] = condition

        return new_slice

    def get_name(self):
        name = 'none_redundant_' + str(self.feature_number) + '_' + str(self.alpha) + '_' + str(self.max_subspace_size)
        return name

    def _maintain_slices(self, generated_slices, data, target, experiment):
        experiment_group = data.groupby(target)[experiment]
        for s1 in generated_slices:

            slice_filter = get_slice_filter(data, s1)

            probabilities = experimentwise_probabilities(slice_filter, data, experiment)
            means = classwise_mean(experiment_group, probabilities)
            variances = classwise_variance(experiment_group, probabilities)
            relevance = build_chebyshev_matrix(means, variances).min()

            s1['probabilities'] = probabilities
            s1['relevance'] = relevance
            s1_redundancies = np.zeros(self.feature_number)

            if np.isinf(s1['relevance']):
                break

            for i in range(len(self.slices)):
                s2 = self.slices[i]
                s1_redundancies[i] = abs(pearsonr(s1['probabilities'], s2['probabilities'])[0])

            # simply add, with feature_number isn't reached yet
            if len(self.slices) < self.feature_number:
                self.redundancies[len(self.slices), :] = s1_redundancies
                self.redundancies[:, len(self.slices)] = s1_redundancies
                self.redundancies.mask = np.identity(self.feature_number)
                self.relevance += s1['relevance']
                self.slices.append(s1)

            # test if slice is beneficial
            else:
                max_score = self.score
                changed = -1
                for i in range(len(self.slices)):
                    red_copy = self.redundancies.copy()
                    red_copy[i, :] = s1_redundancies
                    red_copy[:, i] = s1_redundancies
                    red_copy.mask = np.identity(self.feature_number)
                    red = red_copy.max(axis=1).sum()

                    rel = self.relevance - self.slices[i]['relevance'] + s1['relevance']
                    if max_score < rel-red:
                        max_score = rel-red
                        changed = i

                if changed >= 0:
                    self.redundancies[changed, :] = s1_redundancies
                    self.redundancies[:, changed] = s1_redundancies
                    self.redundancies.mask = np.identity(self.feature_number)
                    self.relevance -= self.slices[changed]['relevance']
                    self.relevance += s1['relevance']
                    self.slices[changed] = s1

            # update score
            self.score = self.relevance-self.redundancies.max(axis=1).sum()

    def fit(self, data, target, experiment, iterations=2000):
        input_fts = [ft for ft in data.columns.values if ft != target and ft != experiment]
        generated_slices = []
        for i in range(iterations):
            subspace_size = randint(1, self.max_subspace_size)
            shuffle(input_fts)
            subspace = input_fts[:subspace_size]

            s = self._generate_slice(data, subspace)
            generated_slices.append(s)

        self._maintain_slices(generated_slices, data, target, experiment)

    def transform(self, data, experiment):
        experiments = data[experiment].unique
        new_features = {}

        for i in range(len(self.slices)):
            s = self.slices[i]
            slice_filter = get_slice_filter(data, s)
            name = json.dumps(s['feature_set']) + str(i)
            new_feature = np.array(experimentwise_probabilities(slice_filter, data, experiment))
            new_features[name] = new_feature

        return pd.DataFrame(data=new_features)


class IncrementalFeatureExtractionMIRedundancy(AbstractFeatureExtraction):

    def __init__(self, feature_number, alpha, max_subspace_size):
        self.alpha = alpha
        self.feature_number = feature_number
        self.max_subspace_size = max_subspace_size
        self.slices = []
        self.cached_values = {}
        self.score = 0
        self.redundancies = np.ma.array(np.zeros((self.feature_number, self.feature_number)),
                                                  mask = np.identity(self.feature_number))
        self.relevance = 0

    def _get_values(self, data, feature):
        if feature not in self.cached_values:
            values, counts = np.unique(data[feature], return_counts=True)
            self.cached_values[feature] = pd.Series(index=values, data=counts)
        return self.cached_values[feature]

    def _generate_slice(self, data, subspace):
        subspace.sort()
        new_slice = {'score': 0, 'feature_set': subspace, 'conditions': {}}
        samples_per_dimension = ceil(pow(self.alpha, 1/len(subspace)) * len(data))

        for ft in subspace:
            ft_values = self._get_values(data, ft)
            shuffled_values = np.random.permutation(ft_values.index.values)
            sample_counter = 0
            condition = []

            for value in shuffled_values:
                if sample_counter >= samples_per_dimension:
                    break
                condition.append(value)
                sample_counter += ft_values[value]
            new_slice['conditions'][ft] = condition

        return new_slice

    def get_name(self):
        name = 'none_redundant_' + str(self.feature_number) + '_' + str(self.alpha) + '_' + str(self.max_subspace_size)
        return name

    def _maintain_slices(self, generated_slices, data, target, experiment):
        experiment_group = data.groupby(target)[experiment]
        for s1 in generated_slices:

            slice_filter = get_slice_filter(data, s1)

            probabilities = experimentwise_probabilities(slice_filter, data, experiment)
            means = classwise_mean(experiment_group, probabilities)
            variances = classwise_variance(experiment_group, probabilities)
            relevance = build_chebyshev_matrix(means, variances).min()

            s1['probabilities'] = probabilities
            s1['relevance'] = relevance
            s1['filter'] = slice_filter
            s1_redundancies = np.zeros(self.feature_number)

            if np.isinf(s1['relevance']):
                break

            for i in range(len(self.slices)):
                s2 = self.slices[i]
                s1_redundancies[i] = mutual_info_score(s1['filter'], s2['filter'])

            # simply add, with feature_number isn't reached yet
            if len(self.slices) < self.feature_number:
                self.redundancies[len(self.slices), :] = s1_redundancies
                self.redundancies[:, len(self.slices)] = s1_redundancies
                self.redundancies.mask = np.identity(self.feature_number)
                self.relevance += s1['relevance']
                self.slices.append(s1)

            # test if slice is beneficial
            else:
                max_score = self.score
                changed = -1
                for i in range(len(self.slices)):
                    red_copy = self.redundancies.copy()
                    red_copy[i, :] = s1_redundancies
                    red_copy[:, i] = s1_redundancies
                    red_copy.mask = np.identity(self.feature_number)
                    red = red_copy.max(axis=1).sum()

                    rel = self.relevance - self.slices[i]['relevance'] + s1['relevance']
                    if max_score < rel-red:
                        max_score = rel-red
                        changed = i

                if changed >= 0:
                    self.redundancies[changed, :] = s1_redundancies
                    self.redundancies[:, changed] = s1_redundancies
                    self.redundancies.mask = np.identity(self.feature_number)
                    self.relevance -= self.slices[changed]['relevance']
                    self.relevance += s1['relevance']
                    self.slices[changed] = s1

            # update score
            self.score = self.relevance-self.redundancies.max(axis=1).sum()

    def fit(self, data, target, experiment, iterations=2000):
        input_fts = [ft for ft in data.columns.values if ft != target and ft != experiment]
        generated_slices = []
        for i in range(iterations):
            subspace_size = randint(1, self.max_subspace_size)
            shuffle(input_fts)
            subspace = input_fts[:subspace_size]

            s = self._generate_slice(data, subspace)
            generated_slices.append(s)

        self._maintain_slices(generated_slices, data, target, experiment)

    def transform(self, data, experiment):
        experiments = data[experiment].unique
        new_features = {}

        for i in range(len(self.slices)):
            s = self.slices[i]
            slice_filter = get_slice_filter(data, s)
            name = json.dumps(s['feature_set']) + str(i)
            new_feature = np.array(experimentwise_probabilities(slice_filter, data, experiment))
            new_features[name] = new_feature

        return pd.DataFrame(data=new_features)



class IncrementalFeatureExtractionKLDRelevance(AbstractFeatureExtraction):

    def __init__(self, feature_number, alpha, max_subspace_size):
        self.alpha = alpha
        self.feature_number = feature_number
        self.max_subspace_size = max_subspace_size
        self.slices = []
        self.cached_values = {}
        self.score = 0
        self.redundancies = np.ma.array(np.zeros((self.feature_number, self.feature_number)),
                                                  mask = np.identity(self.feature_number))
        self.relevance = 0

    def _get_values(self, data, feature):
        if feature not in self.cached_values:
            values, counts = np.unique(data[feature], return_counts=True)
            self.cached_values[feature] = pd.Series(index=values, data=counts)
        return self.cached_values[feature]

    def _generate_slice(self, data, subspace):
        subspace.sort()
        new_slice = {'score': 0, 'feature_set': subspace, 'conditions': {}}
        samples_per_dimension = ceil(pow(self.alpha, 1/len(subspace)) * len(data))

        for ft in subspace:
            ft_values = self._get_values(data, ft)
            shuffled_values = np.random.permutation(ft_values.index.values)
            sample_counter = 0
            condition = []

            for value in shuffled_values:
                if sample_counter >= samples_per_dimension:
                    break
                condition.append(value)
                sample_counter += ft_values[value]
            new_slice['conditions'][ft] = condition

        return new_slice

    def get_name(self):
        name = 'none_redundant_' + str(self.feature_number) + '_' + str(self.alpha) + '_' + str(self.max_subspace_size)
        return name

    def _maintain_slices(self, generated_slices, data, target, experiment):
        experiment_group = data.groupby(target)[experiment]
        for s1 in generated_slices:

            slice_filter = get_slice_filter(data, s1)

            probabilities = experimentwise_probabilities(slice_filter, data, experiment)
            means = classwise_mean(experiment_group, probabilities)
            variances = classwise_variance(experiment_group, probabilities)
            relevance = KLD(data, target, experiment, slice_filter)

            s1['probabilities'] = probabilities
            s1['relevance'] = relevance
            s1_redundancies = np.zeros(self.feature_number)

            if np.isinf(s1['relevance']):
                break

            for i in range(len(self.slices)):
                s2 = self.slices[i]
                s1_redundancies[i] = abs(spearmanr(s1['probabilities'], s2['probabilities'])[0])

            # simply add, with feature_number isn't reached yet
            if len(self.slices) < self.feature_number:
                self.redundancies[len(self.slices), :] = s1_redundancies
                self.redundancies[:, len(self.slices)] = s1_redundancies
                self.redundancies.mask = np.identity(self.feature_number)
                self.relevance += s1['relevance']
                self.slices.append(s1)

            # test if slice is beneficial
            else:
                max_score = self.score
                changed = -1
                for i in range(len(self.slices)):
                    red_copy = self.redundancies.copy()
                    red_copy[i, :] = s1_redundancies
                    red_copy[:, i] = s1_redundancies
                    red_copy.mask = np.identity(self.feature_number)
                    red = red_copy.max(axis=1).sum()

                    rel = self.relevance - self.slices[i]['relevance'] + s1['relevance']
                    if max_score < rel-red:
                        max_score = rel-red
                        changed = i

                if changed >= 0:
                    self.redundancies[changed, :] = s1_redundancies
                    self.redundancies[:, changed] = s1_redundancies
                    self.redundancies.mask = np.identity(self.feature_number)
                    self.relevance -= self.slices[changed]['relevance']
                    self.relevance += s1['relevance']
                    self.slices[changed] = s1

            # update score
            self.score = self.relevance-self.redundancies.max(axis=1).sum()

    def fit(self, data, target, experiment, iterations=2000):
        input_fts = [ft for ft in data.columns.values if ft != target and ft != experiment]
        generated_slices = []
        for i in range(iterations):
            subspace_size = randint(1, self.max_subspace_size)
            shuffle(input_fts)
            subspace = input_fts[:subspace_size]

            s = self._generate_slice(data, subspace)
            generated_slices.append(s)

        self._maintain_slices(generated_slices, data, target, experiment)

    def transform(self, data, experiment):
        experiments = data[experiment].unique
        new_features = {}

        for i in range(len(self.slices)):
            s = self.slices[i]
            slice_filter = get_slice_filter(data, s)
            name = json.dumps(s['feature_set']) + str(i)
            new_feature = np.array(experimentwise_probabilities(slice_filter, data, experiment))
            new_features[name] = new_feature

        return pd.DataFrame(data=new_features)


class IncrementalFeatureExtractionMIRelevance(AbstractFeatureExtraction):

    def __init__(self, feature_number, alpha, max_subspace_size):
        self.alpha = alpha
        self.feature_number = feature_number
        self.max_subspace_size = max_subspace_size
        self.slices = []
        self.cached_values = {}
        self.score = 0
        self.redundancies = np.ma.array(np.zeros((self.feature_number, self.feature_number)),
                                                  mask = np.identity(self.feature_number))
        self.relevance = 0

    def _get_values(self, data, feature):
        if feature not in self.cached_values:
            values, counts = np.unique(data[feature], return_counts=True)
            self.cached_values[feature] = pd.Series(index=values, data=counts)
        return self.cached_values[feature]

    def _generate_slice(self, data, subspace):
        subspace.sort()
        new_slice = {'score': 0, 'feature_set': subspace, 'conditions': {}}
        samples_per_dimension = ceil(pow(self.alpha, 1/len(subspace)) * len(data))

        for ft in subspace:
            ft_values = self._get_values(data, ft)
            shuffled_values = np.random.permutation(ft_values.index.values)
            sample_counter = 0
            condition = []

            for value in shuffled_values:
                if sample_counter >= samples_per_dimension:
                    break
                condition.append(value)
                sample_counter += ft_values[value]
            new_slice['conditions'][ft] = condition

        return new_slice

    def get_name(self):
        name = 'none_redundant_' + str(self.feature_number) + '_' + str(self.alpha) + '_' + str(self.max_subspace_size)
        return name

    def _maintain_slices(self, generated_slices, data, target, experiment):
        experiment_group = data.groupby(target)[experiment]
        for s1 in generated_slices:

            slice_filter = get_slice_filter(data, s1)

            probabilities = experimentwise_probabilities(slice_filter, data, experiment)
            means = classwise_mean(experiment_group, probabilities)
            variances = classwise_variance(experiment_group, probabilities)
            relevance = MI(data, target, experiment, slice_filter)

            s1['probabilities'] = probabilities
            s1['relevance'] = relevance
            s1_redundancies = np.zeros(self.feature_number)

            if np.isinf(s1['relevance']):
                break

            for i in range(len(self.slices)):
                s2 = self.slices[i]
                s1_redundancies[i] = abs(spearmanr(s1['probabilities'], s2['probabilities'])[0])

            # simply add, with feature_number isn't reached yet
            if len(self.slices) < self.feature_number:
                self.redundancies[len(self.slices), :] = s1_redundancies
                self.redundancies[:, len(self.slices)] = s1_redundancies
                self.redundancies.mask = np.identity(self.feature_number)
                self.relevance += s1['relevance']
                self.slices.append(s1)

            # test if slice is beneficial
            else:
                max_score = self.score
                changed = -1
                for i in range(len(self.slices)):
                    red_copy = self.redundancies.copy()
                    red_copy[i, :] = s1_redundancies
                    red_copy[:, i] = s1_redundancies
                    red_copy.mask = np.identity(self.feature_number)
                    red = red_copy.max(axis=1).sum()

                    rel = self.relevance - self.slices[i]['relevance'] + s1['relevance']
                    if max_score < rel-red:
                        max_score = rel-red
                        changed = i

                if changed >= 0:
                    self.redundancies[changed, :] = s1_redundancies
                    self.redundancies[:, changed] = s1_redundancies
                    self.redundancies.mask = np.identity(self.feature_number)
                    self.relevance -= self.slices[changed]['relevance']
                    self.relevance += s1['relevance']
                    self.slices[changed] = s1

            # update score
            self.score = self.relevance-self.redundancies.max(axis=1).sum()

    def fit(self, data, target, experiment, iterations=2000):
        input_fts = [ft for ft in data.columns.values if ft != target and ft != experiment]
        generated_slices = []
        for i in range(iterations):
            subspace_size = randint(1, self.max_subspace_size)
            shuffle(input_fts)
            subspace = input_fts[:subspace_size]

            s = self._generate_slice(data, subspace)
            generated_slices.append(s)

        self._maintain_slices(generated_slices, data, target, experiment)

    def transform(self, data, experiment):
        experiments = data[experiment].unique
        new_features = {}

        for i in range(len(self.slices)):
            s = self.slices[i]
            slice_filter = get_slice_filter(data, s)
            name = json.dumps(s['feature_set']) + str(i)
            new_feature = np.array(experimentwise_probabilities(slice_filter, data, experiment))
            new_features[name] = new_feature

        return pd.DataFrame(data=new_features)


class IncrementalFeatureExtractionAlpha(AbstractFeatureExtraction):

    def __init__(self, feature_number, alpha, max_subspace_size):
        self.alpha = alpha
        self.feature_number = feature_number
        self.max_subspace_size = max_subspace_size
        self.slices = []
        self.cached_values = {}
        self.score = 0
        self.redundancies = np.ma.array(np.zeros((self.feature_number, self.feature_number)),
                                                  mask = np.identity(self.feature_number))
        self.relevance = 0

    def _get_values(self, data, feature):
        if feature not in self.cached_values:
            values, counts = np.unique(data[feature], return_counts=True)
            self.cached_values[feature] = pd.Series(index=values, data=counts)
        return self.cached_values[feature]

    def _generate_slice(self, data, subspace):
        subspace.sort()
        new_slice = {'score': 0, 'feature_set': subspace, 'conditions': {}}
        alpha = uniform(0.1, 0.9)
        samples_per_dimension = ceil(pow(alpha, 1/len(subspace)) * len(data))

        for ft in subspace:
            ft_values = self._get_values(data, ft)
            shuffled_values = np.random.permutation(ft_values.index.values)
            sample_counter = 0
            condition = []

            for value in shuffled_values:
                if sample_counter >= samples_per_dimension:
                    break
                condition.append(value)
                sample_counter += ft_values[value]
            new_slice['conditions'][ft] = condition

        return new_slice

    def get_name(self):
        name = 'none_redundant_' + str(self.feature_number) + '_' + str(self.alpha) + '_' + str(self.max_subspace_size)
        return name

    def _maintain_slices(self, generated_slices, data, target, experiment):
        experiment_group = data.groupby(target)[experiment]
        for s1 in generated_slices:

            slice_filter = get_slice_filter(data, s1)

            probabilities = experimentwise_probabilities(slice_filter, data, experiment)
            means = classwise_mean(experiment_group, probabilities)
            variances = classwise_variance(experiment_group, probabilities)
            relevance = build_chebyshev_matrix(means, variances).min()

            s1['probabilities'] = probabilities
            s1['relevance'] = relevance
            s1_redundancies = np.zeros(self.feature_number)

            if np.isinf(s1['relevance']):
                break

            for i in range(len(self.slices)):
                s2 = self.slices[i]
                s1_redundancies[i] = abs(spearmanr(s1['probabilities'], s2['probabilities'])[0])

            # simply add, with feature_number isn't reached yet
            if len(self.slices) < self.feature_number:
                self.redundancies[len(self.slices), :] = s1_redundancies
                self.redundancies[:, len(self.slices)] = s1_redundancies
                self.redundancies.mask = np.identity(self.feature_number)
                self.relevance += s1['relevance']
                self.slices.append(s1)

            # test if slice is beneficial
            else:
                max_score = self.score
                changed = -1
                for i in range(len(self.slices)):
                    red_copy = self.redundancies.copy()
                    red_copy[i, :] = s1_redundancies
                    red_copy[:, i] = s1_redundancies
                    red_copy.mask = np.identity(self.feature_number)
                    red = red_copy.max(axis=1).sum()

                    rel = self.relevance - self.slices[i]['relevance'] + s1['relevance']
                    if max_score < rel-red:
                        max_score = rel-red
                        changed = i

                if changed >= 0:
                    self.redundancies[changed, :] = s1_redundancies
                    self.redundancies[:, changed] = s1_redundancies
                    self.redundancies.mask = np.identity(self.feature_number)
                    self.relevance -= self.slices[changed]['relevance']
                    self.relevance += s1['relevance']
                    self.slices[changed] = s1

            # update score
            self.score = self.relevance-self.redundancies.max(axis=1).sum()

    def fit(self, data, target, experiment, iterations=2000):
        input_fts = [ft for ft in data.columns.values if ft != target and ft != experiment]
        generated_slices = []
        for i in range(iterations):
            subspace_size = randint(1, self.max_subspace_size)
            shuffle(input_fts)
            subspace = input_fts[:subspace_size]

            s = self._generate_slice(data, subspace)
            generated_slices.append(s)

        self._maintain_slices(generated_slices, data, target, experiment)

    def transform(self, data, experiment):
        experiments = data[experiment].unique
        new_features = {}

        for i in range(len(self.slices)):
            s = self.slices[i]
            slice_filter = get_slice_filter(data, s)
            name = json.dumps(s['feature_set']) + str(i)
            new_feature = np.array(experimentwise_probabilities(slice_filter, data, experiment))
            new_features[name] = new_feature

        return pd.DataFrame(data=new_features)
