import numpy as np
import pandas as pd
from scipy.optimize import minimize_scalar
from sklearn.metrics import mutual_info_score


# helper

def get_slice_filter(data, s):
    slice_filter = np.ones(len(data))
    for ft, values in s['conditions'].items():
        slice_filter = np.logical_and(slice_filter, data[ft].isin(values))
    return pd.Series(slice_filter)


def experimentwise_probabilities(slice_filter, data, experiment):
    return slice_filter.groupby(data[experiment]).apply(lambda x: x.sum()/x.count())


def classwise_mean(experiment_by_target, prob_per_experiment):
    return experiment_by_target.apply(lambda x: prob_per_experiment[x.unique()].mean())


def classwise_variance(experiment_by_target, prob_per_experiment):
    return experiment_by_target.apply(lambda x: prob_per_experiment[x.unique()].var())


def _error_bound(var1, var2, dist, x):
    return var1/2*x**2 + var2/2*(dist-x)**2


def _optimized_error_bound(var1, var2, dist):
    if dist == 0:
        return np.inf
    return minimize_scalar(lambda x: _error_bound(var1, var2, dist, x), bounds=(0, dist), method='bounded').fun


_vectorized_optimized_error_bound = np.vectorize(_optimized_error_bound)


def build_chebyshev_matrix(means, variances):
    mean_matrix = np.array([means.values]*len(means))
    mean_distances = (np.absolute(mean_matrix - mean_matrix.T)**2) + 1/(len(means)*1e6)

    variance_matrix = np.array([variances.values]*len(variances))
    variance_sums = variance_matrix + variance_matrix.T

    distance_zero = (mean_distances == 0)
    variance_sums[distance_zero] = np.inf
    mean_distances[distance_zero] = 1

    score_matrix = 1 - variance_sums/mean_distances

    np.fill_diagonal(score_matrix, 1)
    return score_matrix


def build_minimized_chebyshev_matrix(means, variances):
    mean_matrix = np.array([means.values]*len(means))
    mean_distances = np.absolute(mean_matrix - mean_matrix.T)**2

    variance_matrix = np.array([variances.values]*len(variances))

    optimized_error = _vectorized_optimized_error_bound(variance_matrix,
                                                        variance_matrix.T,
                                                        mean_distances)

    np.fill_diagonal(optimized_error, 0)

    result = 1 - optimized_error
    return result


# scoring

def chebyshev_bound_minimized(data, target, experiment, slice_filter):
    prob_experiment = experimentwise_probabilities(slice_filter, data, experiment)
    experiment_group = data.groupby(target)[experiment]

    probabilities = classwise_mean(experiment_group, prob_experiment)
    variances = classwise_variance(experiment_group, prob_experiment)

    result = build_minimized_chebyshev_matrix(probabilities, variances)
    return result.min()


def chebyshev_bound(data, target, experiment, slice_filter):
    prob_experiment = experimentwise_probabilities(slice_filter, data, experiment)
    experiment_group = data.groupby(target)[experiment]

    probabilities = classwise_mean(experiment_group, prob_experiment)
    variances = classwise_variance(experiment_group, prob_experiment)

    score_matrix = build_chebyshev_matrix(probabilities, variances)
    return score_matrix.min()


def min_probability_distance_experimentwise(data, target, experiment, slice_filter):
    prob_experiment = slice_filter.groupby(data[experiment]).apply(_probabilities)
    experiment_group = data.groupby(target)[experiment]

    probabilities = classwise_mean(experiment_group, prob_experiment)
    probability_matrix = np.array([probabilities.values]*len(probabilities))
    probability_distances = np.absolute(probability_matrix - probability_matrix.T)
    np.fill_diagonal(probability_distances, np.inf)
    return probability_distances.min()


def min_probability_distance(data, target, experiment, slice_filter):
    probabilities = slice_filter.groupby(data[target]).apply(lambda x: x.sum()/x.count())
    probability_matrix = np.array([probabilities.values]*len(probabilities))
    probability_distances = np.absolute(probability_matrix - probability_matrix.T)
    np.fill_diagonal(probability_distances, np.inf)
    return probability_distances.min()


def KLD(data, target, experiment, slice_filter):
    marginal_classes, marginal_counts = np.unique(data[target], return_counts=True)
    marginal_distribution = pd.Series(index=marginal_classes, data=marginal_counts/marginal_counts.sum())

    conditional_classes, conditional_counts = np.unique(data.loc[slice_filter==1, target], return_counts=True)
    conditional_distribution = pd.Series(index=conditional_classes, data=conditional_counts/conditional_counts.sum())

    marginal_distribution = marginal_distribution[conditional_distribution.index]
    return (conditional_distribution * np.log2(conditional_distribution/marginal_distribution)).sum()


def JSD(data, target, experiment, slice_filter):
    slice_probability = slice_filter.sum()/slice_filter.count()
    kld_slice = KLD(data, target, experiment, slice_filter)
    kld_not_slice = KLD(data, target, experiment, slice_filter!=1)
    return slice_probability*kld_slice + (1 - slice_probability)*kld_not_slice


def MI(data, target, experiment, slice_filter):
    target_labels = data[target]
    return mutual_info_score(slice_filter, target_labels)